@@ -11,6 +11,7 @@ use std::path::PathBuf;
 use crate::ViewfinderState;
 
 const BARCODE_TIMEOUT: u32 = 1;
+const PROVIDER_TIMEOUT: u64 = 2;
 
 mod imp {
     use super::*;
@@ -23,7 +24,7 @@ mod imp {
     #[derive(Debug, Default, Properties)]
     #[properties(wrapper_type = super::Viewfinder)]
     pub struct Viewfinder {
-        #[property(get, set = Self::set_state, explicit_notify, builder(Default::default()))]
+        #[property(get, explicit_notify, builder(Default::default()))]
         state: Cell<ViewfinderState>,
         #[property(get = Self::detect_codes, set = Self::set_detect_codes, explicit_notify)]
         detect_codes: Cell<bool>,
@@ -45,7 +46,7 @@ mod imp {
 
         pub timeout_handler: RefCell<Option<glib::SourceId>>,
 
-        picture: gtk::Picture,
+        pub picture: gtk::Picture,
     }
 
     impl Viewfinder {
@@ -53,7 +54,7 @@ mod imp {
             self.camerabin.get().unwrap()
         }
 
-        fn set_state(&self, state: ViewfinderState) {
+        pub(crate) fn set_state(&self, state: ViewfinderState) {
             if state != self.state.replace(state) {
                 self.obj().notify_state();
             }
@@ -90,7 +91,7 @@ mod imp {
         fn set_camera(&self, camera: Option<crate::Camera>) {
             let obj = self.obj();
 
-            if !matches!(obj.state(), ViewfinderState::Ready) {
+            if !matches!(obj.state(), ViewfinderState::Ready | ViewfinderState::Error) {
                 log::error!("Could not set camera, the viewfinder is not ready");
                 return;
             }
@@ -109,7 +110,28 @@ mod imp {
                 return;
             }
 
-            if obj.is_realized() && self.camerabin().current_state() == gst::State::Playing {
+            // We reset to READY if we landed on the ERROR state on the previous
+            // camera.
+            if matches!(obj.state(), ViewfinderState::Error) {
+                if self
+                    .devices
+                    .get()
+                    .and_then(|devices| devices.camera(0))
+                    .is_some()
+                {
+                    self.set_state(ViewfinderState::Ready);
+                } else {
+                    self.set_state(ViewfinderState::NoCameras);
+                }
+            }
+
+            // The current state is PAUSED if there was an error on the previous camera.
+            if obj.is_realized()
+                && matches!(
+                    self.camerabin().current_state(),
+                    gst::State::Playing | gst::State::Paused
+                )
+            {
                 obj.stop_camerabin();
             }
 
@@ -229,8 +251,8 @@ mod imp {
             let devices = crate::DeviceProvider::instance();
 
             devices.connect_camera_added(glib::clone!(@weak obj => move |_, camera| {
-                if matches!(obj.state(), ViewfinderState::NoCameras) {
-                    obj.set_state(ViewfinderState::Ready);
+                if matches!(obj.state(), ViewfinderState::NoCameras | ViewfinderState::Loading | ViewfinderState::Error) {
+                    obj.imp().set_state(ViewfinderState::Ready);
                     obj.set_camera(Some(camera.clone()));
                 }
             }));
@@ -244,22 +266,29 @@ mod imp {
                     let is_none = next_camera.is_none();
                     obj.set_camera(next_camera);
                     if is_none {
-                        obj.set_state(ViewfinderState::NoCameras);
+                        obj.imp().set_state(ViewfinderState::NoCameras);
                     }
                 }
             }));
 
             if let Some(camera) = devices.camera(0) {
-                obj.set_state(ViewfinderState::Ready);
+                obj.imp().set_state(ViewfinderState::Ready);
                 obj.set_camera(Some(camera));
-            } else {
-                obj.set_state(ViewfinderState::NoCameras);
             }
 
             self.devices.set(devices.clone()).unwrap();
 
             log::debug!("Setup recording");
             obj.setup_recording();
+
+            glib::timeout_add_local_once(
+                std::time::Duration::from_secs(PROVIDER_TIMEOUT),
+                glib::clone!(@weak obj => move || {
+                    if matches!(obj.state(), ViewfinderState::Loading) {
+                        obj.imp().set_state(ViewfinderState::NoCameras);
+                    }
+                }),
+            );
         }
 
         fn dispose(&self) {
@@ -394,6 +423,10 @@ glib::wrapper! {
     /// This will only be emitted if [`detect-codes`](#detect-codes) is `true`.
     ///
     /// Barcodes are only detected when they appear on the feed, not on every frame when they are visible.
+    ///
+    /// # Implements
+    ///
+    /// [`gtk::prelude::WidgetExt`][trait@gtk::prelude::WidgetExt], [`glib::ObjectExt`][trait@gtk::glib::ObjectExt]
     pub struct Viewfinder(ObjectSubclass<imp::Viewfinder>)
         @extends gtk::Widget;
 }
@@ -414,6 +447,20 @@ impl Viewfinder {
         Self::default()
     }
 
+    /// Gets the aspect ratio of the camera output.
+    ///
+    /// # Returns
+    ///
+    /// an aspect ratio calculated with width/height, or 0 for no valid aspect ratio.
+    pub fn aspect_ratio(&self) -> f64 {
+        let imp = self.imp();
+        if let Some(paintable) = imp.picture.paintable() {
+            return paintable.intrinsic_aspect_ratio();
+        } else {
+            return 0.0;
+        }
+    }
+
     /// Takes a picture.
     ///
     /// The recording will be saved to `location`. This method throws an error if:
@@ -556,7 +603,7 @@ impl Viewfinder {
         // TODO Present a toast, banner, or message dialog with the error.
         if let Err(err) = self.imp().camerabin().set_state(gst::State::Playing) {
             log::error!("Could not start camerabin: {err}");
-            self.set_state(ViewfinderState::Error);
+            self.imp().set_state(ViewfinderState::Error);
         } else {
             log::debug!("Camerabin state succesfully set to PLAYING");
         }
@@ -565,7 +612,7 @@ impl Viewfinder {
     fn stop_camerabin(&self) {
         if let Err(err) = self.imp().camerabin().set_state(gst::State::Null) {
             log::error!("Could not pause camerabin: {err}");
-            self.set_state(ViewfinderState::Error);
+            self.imp().set_state(ViewfinderState::Error);
         } else {
             log::debug!("Camerabin state succesfully set to NULL");
         }
@@ -637,7 +684,7 @@ impl Viewfinder {
         self.cancel_current_operation();
 
         if self.imp().camerabin().current_state() != gst::State::Playing {
-            self.set_state(ViewfinderState::Error);
+            self.imp().set_state(ViewfinderState::Error);
         }
     }
 
